#ifndef WIN32
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <math.h>
#include <string.h>
#include "AudioPort.h"
#endif
#include "AudioFile.h"
#include "g711.h"
#include "fft.h"


#if defined Linux || defined WIN32 || defined _LITTLE_ENDIAN
#  define LE
#endif

Audio::AFfilehandle::AFfilehandle(int _sampfreq, int _sampformat, int _nchannels)
{
  sampfreq    = _sampfreq;
  sampformat  = _sampformat;
  if (sampformat == LIN16)
    sampsize  = 2;
  else
    sampsize  = 1;
  nchannels = _nchannels;

  maxsamp     = 0;
  minsamp     = 0;
  length      = 0;
  maxlength   = 0;
  abmax       = 0;
  active      = IDLE;
  swap        = 0;
  inByteOrder = NATIVE;
}

Audio::AFfilehandle::~AFfilehandle()
{
}


Audio::File::File(int _sampfreq, int _sampformat, int _nchannels)
{
  filehandle = new AFfilehandle(_sampfreq,_sampformat,_nchannels);
  filetype = NULL;
}

Audio::File::~File()
{
  delete filehandle;
  if (filename) free(filename);
  if (filetype) free(filetype);
}


int Audio::File::Open(char *_filename, char *mode)
{
  if (_filename) {
    filename = strdup(_filename);
    fp = fopen(filename,mode);
  }
  else {
    filename = NULL;
    fp = NULL;
    return 0;  
  }

  return 1;
}

int Audio::File::Close()
{
  if (fp) fclose(fp);
  fp = NULL;

  return 1;
}

int Audio::File::Seek(int pos)
{
  if (fp) fseek(fp,pos,SEEK_SET);

  return 1;
}

char *Audio::File::GuessWav(char *buf, int len)
{
  if (len < 21) return(QUE_STRING);
  if (strncmp("RIFF", buf, strlen("RIFF")) == 0) {
    if (strncmp("WAVE", &buf[8], strlen("WAVE")) == 0) {
      return(WAV_STRING);
    }
  }
  return(NULL);
}

char *Audio::File::GuessAu(char *buf, int len)
{
  if (len < 4) return(QUE_STRING);
  if (strncmp(".snd", buf, strlen(".snd")) == 0) {
    return(AU_STRING);
  }
  return(NULL);
}

char *Audio::File::GuessAiff(char *buf, int len)
{
  if (len < 20) return(QUE_STRING);
  if (strncmp("FORM", buf, strlen("FORM")) == 0) {
    if (strncmp("AIFF", &buf[8], strlen("AIFF")) == 0) {
      return(AIFF_STRING);
    }
  }
  return(NULL);
}

int Audio::File::GuessFileType(char *buf, int len, int eof)
{
  int 
    flag = 0;

  if (filetype) free(filetype);

  if (filetype = GuessWav(buf, len));
  else if (filetype = GuessAiff(buf,len));
  else if (filetype = GuessAu(buf,len));
  
  if (filetype == NULL) {
    /* guessProc can't recognize this header */
  } else if (strcmp(filetype, QUE_STRING) == 0) {
    flag = 1; /* guessProc needs more bytes in order to decide */
  } else if (strcmp(filetype, RAW_STRING) != 0) {
    return 1;
  }
  
  /* Don't decide yet if there's more header bytes to be had */

  if (flag && !eof) {
    return -1;
  }

  /* No guessProc recognized this header => guess RAW format */
  
  return -1;
}

void Audio::File::GetExtremes(int start, int end, int *pmax, int *pmin)
{
  int 
    i, inc,
    maxs, mins;

  if (filehandle->length == 0) {
    if (filehandle->sampformat == LIN8OFFSET) {
      *pmax = 128;
      *pmin = 128;
    } 
    else {
      *pmax = 0;
      *pmin = 0;
    }
    return;
  }
  
  inc = filehandle->nchannels;
  start = start * filehandle->nchannels;
  end   = end * filehandle->nchannels;

  if (filehandle->sampformat == LIN8OFFSET) {
    maxs = 0;
    mins = 255;
  } 
  else if (filehandle->sampformat == LIN8) {
    maxs = -128;
    mins = 127;
  } 
  else {
    maxs = -32768;
    mins = 32767;
  }

  switch (filehandle->sampformat) {
  case LIN16:
    for (i = start; i <= end; i += inc) {
      short tmp = GetSample(i);
      if (tmp > maxs) {
	maxs = tmp;
      }
      if (tmp < mins) {
	mins = tmp;
      }
    }
    break;
  case ALAW:
    for (i = start; i <= end; i += inc) {
      int tmp = G711::Alaw2Lin((unsigned char)GetSample(i));
      if (tmp > maxs) maxs = tmp;
      if (tmp < mins) mins = tmp;
    }
    break;
  case MULAW:
    for (i = start; i <= end; i += inc) {
      int tmp = G711::Mulaw2Lin((unsigned char)GetSample(i));
      if (tmp > maxs) maxs = tmp;
      if (tmp < mins) mins = tmp;
    }
    break;
  case LIN8OFFSET:
    for (i = start; i <= end; i += inc) {
      unsigned char tmp = (unsigned char)GetSample(i);
      if (tmp > maxs) maxs = tmp;
      if (tmp < mins) mins = tmp;
    }
    break;
  default:
    for (i = start; i <= end; i += inc) {
      char tmp = (char )GetSample(i);
      if (tmp > maxs) maxs = tmp;
      if (tmp < mins) mins = tmp;
    }
  }

  if (maxs < mins) {
    maxs = mins;
  }
  if (mins > maxs) {
    mins = maxs;
  }

  *pmax = maxs;
  *pmin = mins;
}

void Audio::File::UpdateExtremes(int start, int end, int flag)
{
  int 
    maxs, mins, 
    newmax, newmin;
  
  maxs = filehandle->maxsamp;
  mins = filehandle->minsamp;

  GetExtremes(start, end - 1, &newmax, &newmin);

  if (newmax > maxs) {
    filehandle->maxsamp = newmax;
  } 
  else {
    filehandle->maxsamp = maxs;
  }
  if (newmin < mins) {
    filehandle->minsamp = newmin;
  } 
  else {
    filehandle->minsamp = mins;
  }
  if (filehandle->maxsamp > -filehandle->minsamp)
    filehandle->abmax = filehandle->maxsamp;
  else
    filehandle->abmax = -filehandle->minsamp;
}

short Audio::File::GetSample(int index)
{
  int 
    nRead, 
    size = ITEMBUFFERSIZE, i;

  if (filehandle->active == READ) {
    return SSAMPLE(filehandle->tmpbuf, index);
  }

  if (index < filehandle->filePos + ITEMBUFFERSIZE / filehandle->sampsize
      && index >= filehandle->filePos && filehandle->filePos != -1) {
    if (filehandle->sampsize == 2) {
      return(((short *)filehandle->buffer)[index-filehandle->filePos]);
    } else if (filehandle->sampformat == LIN8OFFSET) {
      return(filehandle->buffer[index-filehandle->filePos]);
    } else /*if (filehandle->sampformat == LIN8)*/ {
      return(((char *)filehandle->buffer)[index-filehandle->filePos]);
    }
  } else {
    int pos = 0;
    Seek(filehandle->headSize + index * filehandle->sampsize);
    nRead = fread((char *)filehandle->buffer, sizeof(char), size, fp);
    if (filehandle->sampsize == 2 && filehandle->swap)
      for (i = 0; i < nRead/2; i++)
	((short *)filehandle->buffer)[i] =
	  SwapShort(((short *)filehandle->buffer)[i]);
    filehandle->filePos = index;
    if (filehandle->sampsize == 2) {
      return(((short *)filehandle->buffer)[pos]);
    } else {
      return(filehandle->buffer[pos]);
    }
  }
}

short Audio::File::SwapShort(short s)
{
  char 
    tc, *p;

  p = (char *) &s;
  tc = *p;
  *p = *(p+1);
  *(p+1) = tc;
  
  return(s);
}

long Audio::File::SwapLong(long l)
{
  char 
    tc, *p;

  p = (char *) &l;
  tc = *p;
  *p = *(p+3);
  *(p+3) = tc;

  tc = *(p+1);
  *(p+1) = *(p+2);
  *(p+2) = tc;
  
  return(l);
}

void Audio::File::ByteSwapSound()
{
  int 
    i, j;

  for (j = 0; j < filehandle->nblks; j++)
    for (i = 0; i < min(SBLKSIZE, filehandle->length * filehandle->nchannels); i++)
      filehandle->blocks[j][i] = SwapShort(filehandle->blocks[j][i]);
}

void Audio::File::SwapIfBE()
{
#ifdef LE
  filehandle->swap = 0;
#else
  filehandle->swap = 1;
#endif
}

void Audio::File::SwapIfLE()
{
#ifdef LE
  filehandle->swap = 1;
#else
  filehandle->swap = 0;
#endif
}

short Audio::File::ReadBEShort()
{
  short 
    ts;

  fread ((char *) &ts, sizeof(short), 1, fp);

#ifdef LE
  ts = SwapShort(ts);
#endif

  return(ts);
}

short Audio::File::ReadLEShort()
{
  short 
    ts;

  fread((char *) &ts, sizeof(short), 1, fp);

#ifndef LE
  ts = SwapShort(ts);
#endif

  return(ts);
}

long Audio::File::ReadBELong()
{
  long 
    tl;

  fread((char *) &tl, sizeof(long), 1, fp);

#ifdef LE
  tl = SwapLong(tl);
#endif

  return(tl);
}

long Audio::File::ReadLELong()
{
  long 
    tl;

  fread((char *) &tl, sizeof(long), 1, fp);

#ifndef LE
  tl = SwapLong(tl);
#endif

  return(tl);
}

#ifdef WIN32

int Audio::File::Write(char *buf, int len)
{  
  int 
    i = 0;
  short 
    *p;
  char
    *q,
		c1,c2;

  if (filehandle->sampsize == 2) {
    if (filehandle->inByteOrder == NATIVE && filehandle->swap) {
#ifdef LE
      filehandle->inByteOrder = BIGENDIAN;
#else
      filehandle->inByteOrder = LITTLEENDIAN;
#endif
    }

		q = (char *)buf;
    p = (short *)buf;
    for (i = 0; i < len; i++) {
    switch (filehandle->inByteOrder) {
      case NATIVE: 
				c1 = *q;
				q++;
				c2 = *q;
				q++;
				if (fwrite(&c2, 1, 1, fp) == -1) return -1;
				if (fwrite(&c1, 1, 1, fp) == -1) return -1;
				
			//	if (fwrite(&c1, 1, 1, fp) == -1) return -1;
				break;
      case BIGENDIAN:
				if (WriteBEShort(*p) == -1) return -1;
				break;
      case LITTLEENDIAN: 
				if (WriteLEShort((short)buf[i]) == -1) return -1;
				break;
      }
    }
  } 
  else {
    q = (char *)buf;
    for (i = 0; i < len; i++) {
			*q = (*q)^128;
      q++;
    }
    if ((i = fwrite(buf, 1, len, fp)) == -1) {
      return -1;
    }
 //   if (fflush(fp) == EOF) return -1;
  }

	return i;
}

#else

int Audio::File::Write(short *buf, int len)
{  
  int 
    i = 0, j;
  short 
    sh;
  short 
    *p;
  char
    *q;
  unsigned char
    c;


  if (filehandle->sampsize == 2) {
    if (filehandle->inByteOrder == NATIVE && filehandle->swap) {
#ifdef LE
      filehandle->inByteOrder = BIGENDIAN;
#else
      filehandle->inByteOrder = LITTLEENDIAN;
#endif
    }
    p = buf;
    for (i = 0; i < len; i++) {
      switch (filehandle->inByteOrder) {
      case NATIVE:    
        sh = *p;
	if (fwrite((char *) &sh, sizeof(short), 1, fp) == -1) return 0;
	break;
      case BIGENDIAN:
	if (WriteBEShort(*p) == -1) return 0;
	break;
      case LITTLEENDIAN:
	if (WriteLEShort(*p) == -1) return 0;
	break;
      }
      p++;
    }
  } 
  else {
    q = (char *)buf;
    for (i = 0; i < len; i++) {
      c = *q;
      if (fwrite(q, 1, 1, fp) == -1) return 0;
      q++;
    }
  }
//   else {
//     for (i = 0; i < len; i++) {
//       sh = *p;
//       if (fwrite((char *) &sh, 1, 1, fp) == -1) return -1;
//       p++;
//     }
//   }

}

#endif

int Audio::File::WriteLEShort(short s)
{
  short ts = s;

#ifndef LE
  ts = SwapShort(ts);
#endif

  return(fwrite((void *)&ts, sizeof(short), 1, fp));
}

int Audio::File::WriteLELong(long l)
{
  long tl = l;

#ifndef LE
  tl = SwapLong(tl);
#endif

  return(fwrite((void *) &tl, sizeof(long), 1, fp));
}

int Audio::File::WriteBEShort(short s)
{
  short ts = s;

#ifdef LE
  ts = SwapShort(ts);
#endif

  return(fwrite((char *) &ts, sizeof(short), 1, fp));
}

int Audio::File::WriteBELong(long l)
{
  long tl = l;

#ifdef LE
  tl = SwapLong(tl);
#endif

  return(fwrite((char *) &tl, sizeof(long), 1, fp));
}
  
long Audio::File::GetLELong(char *buf, int pos)
{
  long tl;

  memcpy(&tl, &buf[pos], sizeof(long));

#ifndef LE
  tl = SwapLong(tl);
#endif

  return(tl);
}

short Audio::File::GetLEShort(char *buf, int pos)
{
  short ts;
  char *p;
  short *q;

  p = &buf[pos];
  q = (short *) p;
  ts = *q;

#ifndef LE
  ts = SwapShort(ts);
#endif

  return(ts);
}

long Audio::File::GetBELong(char *buf, int pos)
{
  long tl;

  memcpy(&tl, &buf[pos], sizeof(long));

#ifdef LE
  tl = SwapLong(tl);
#endif

  return(tl);
}

short Audio::File::GetBEShort(char *buf, int pos)
{
  short ts;
  char *p;
  short *q;

  p = &buf[pos];
  q = (short *) p;
  ts = *q;

#ifdef LE
  ts = SwapShort(ts);
#endif

  return(ts);
}

void Audio::File::PutBELong(char *buf, int pos, long l)
{
  long tl = l;

#ifdef LE
  tl = SwapLong(tl);
#endif

  memcpy(&buf[pos], &tl, sizeof(long));
}

void Audio::File::PutBEShort(char *buf, int pos, short s)
{
  short ts = s;
  char *p;
  short *q;

  p = &buf[pos];
  q = (short *) p;

#ifdef LE
  ts = SwapShort(ts);
#endif

  *q = ts;
}

/* Note: pos must be a multiple of 4 */

void Audio::File::PutLELong(char *buf, int pos, long l)
{
  long tl = l;
  char *p;
  long *q;

  p = &buf[pos];
  q = (long *) p;

#ifndef LE
  tl = SwapLong(tl);
#endif

  *q = tl;
}

void Audio::File::PutLEShort(char *buf, int pos, short s)
{
  short ts = s;
  char *p;
  short *q;

  p = &buf[pos];
  q = (short *) p;

#ifndef LE
  ts = SwapShort(ts);
#endif

  *q = ts;
}



int Audio::File::ExtCmp(char *s1, char *s2)
{
  int l1 = strlen(s1);
  int l2 = strlen(s2);

  return(strncmp(s1, &s2[l2 - l1], l1));
}

char *Audio::File::ExtWavFile(char *s)
{
  if (ExtCmp(".wav", s) == 0) {
    return(WAV_STRING);
  }
  return(NULL);
}

char *Audio::File::ExtAuFile(char *s)
{
  if (ExtCmp(".au", s) == 0 || ExtCmp(".snd", s) == 0) {
    return(AU_STRING);
  }
  return(NULL);
}

char *Audio::File::ExtAiffFile(char *s)
{
  if (ExtCmp(".aif", s) == 0 || ExtCmp(".aiff", s) == 0) {
    return(AIFF_STRING);
  }
  return(NULL);
}

int Audio::File::GuessFormat(unsigned char *buf, int len) {
  int 
    i, j, 
    format;
  float 
    energyLIN16 = 0.0, 
    energyLIN16S = 0.0,
    energyMULAW = 0.0, 
    energyALAW = 0.0,
    energyLIN8  = 0.0, 
    energyLIN8O = 0.0, 
    minEnergy,
    fft[GUESS_FFT_LENGTH],
    totfft[GUESS_FFT_LENGTH],
    hamwin[GUESS_FFT_LENGTH];
  double 
    toterg = 0.0, 
    cmperg = 0.0, 
    minBin = 0.0;

  /*
    Byte order and sample encoding detection suggested by David van Leeuwen
    */
  
  for (i = 0; i < len / 2; i++) {
    short sampleLIN16  = ((short *)buf)[i];
    short sampleLIN16S = SwapShort(sampleLIN16);
    short sampleMULAW  = G711::Mulaw2Lin(buf[i]);
    short sampleALAW   = G711::Alaw2Lin(buf[i]);
    short sampleLIN8O  = (char)(buf[i] ^ 128) << 8;
    short sampleLIN8   = (char)buf[i] << 8;

    energyLIN16  += (float) sampleLIN16  * (float) sampleLIN16;
    energyLIN16S += (float) sampleLIN16S * (float) sampleLIN16S;
    energyMULAW  += (float) sampleMULAW  * (float) sampleMULAW;
    energyALAW   += (float) sampleALAW   * (float) sampleALAW;
    energyLIN8O  += (float) sampleLIN8O  * (float) sampleLIN8O;
    energyLIN8   += (float) sampleLIN8   * (float) sampleLIN8;
  }
  
  format = GUESS_LIN16;
  minEnergy = energyLIN16;
  
  if (energyLIN16S < minEnergy) {
    format = GUESS_LIN16S;
    minEnergy = energyLIN16S;
  }
  if (energyALAW < minEnergy) {
    format = GUESS_ALAW;
    minEnergy = energyALAW;
  }
  if (energyMULAW < minEnergy) {
    format = GUESS_MULAW;
    minEnergy = energyMULAW;
  }
  if (energyLIN8O < minEnergy) {
    format = GUESS_LIN8OFFSET;
    minEnergy = energyLIN8O;
  }
  if (energyLIN8 < minEnergy) {
    format = GUESS_LIN8;
    minEnergy = energyLIN8;
  }
  
  switch (format) {
  case GUESS_LIN16:
    filehandle->swap = 0;
    if (filehandle->sampsize == 1) {
      filehandle->length /= 2;
    }
    filehandle->sampformat = LIN16;
    filehandle->sampsize = 2;
    break;
  case GUESS_LIN16S:
    filehandle->swap = 1;
    if (filehandle->sampsize == 1) {
      filehandle->length /= 2;
    }
    filehandle->sampformat = LIN16;
    filehandle->sampsize = 2;
    break;
  case GUESS_ALAW:
    if (filehandle->sampsize == 2) {
      filehandle->length *= 2;
    }
    filehandle->sampformat = ALAW;
    filehandle->sampsize = 1;
    filehandle->sampfreq = DEFAULT_ALAW_FREQ;
    break;
  case GUESS_MULAW:
    if (filehandle->sampsize == 2) {
      filehandle->length *= 2;
    }
    filehandle->sampformat = MULAW;
    filehandle->sampsize = 1;
    filehandle->sampfreq = DEFAULT_MULAW_FREQ;
    break;
  case GUESS_LIN8OFFSET:
    if (filehandle->sampsize == 2) {
      filehandle->length *= 2;
    }
    filehandle->sampformat = LIN8OFFSET;
    filehandle->sampsize = 1;
    filehandle->sampfreq = DEFAULT_LIN8OFFSET_FREQ;
    break;
  case GUESS_LIN8:
    if (filehandle->sampsize == 2) {
      filehandle->length *= 2;
    }
    filehandle->sampformat = LIN8;
    filehandle->sampsize = 1;
    filehandle->sampfreq = DEFAULT_LIN8_FREQ;
    break;
  }

  if (filehandle->sampformat == LIN16) {
    for (i = 0; i < GUESS_FFT_LENGTH; i++) {
      totfft[i] = 0.0;
    }
    FFT::Init(GUESS_FFT_LENGTH);
    for (i = 0; i < (len / filehandle->sampsize) / (GUESS_FFT_LENGTH + 1); i++) {
      for (j = 0; j < GUESS_FFT_LENGTH; j++) {
	short sample  = ((short *)buf)[j + i * (GUESS_FFT_LENGTH / 2)];
	if (filehandle->swap) {
	  sample = SwapShort(sample);
	}
	fft[j] = (float) sample * hamwin[j];
      }
      FFT::DBPowerSpectrum(fft);
      for (j = 0; j < GUESS_FFT_LENGTH / 2; j++) {
	totfft[j] += fft[j];
      }
    }
    for (i = 0; i < GUESS_FFT_LENGTH / 2; i++) {
      if (totfft[i] < minBin) minBin = totfft[i];
    }
    for (i = 0; i < GUESS_FFT_LENGTH / 2; i++) {
      toterg += (totfft[i] - minBin);
    }
    for (i = 0; i < GUESS_FFT_LENGTH / 2; i++) {
      cmperg += (totfft[i] - minBin);
      if (cmperg > toterg / 2.0) break;
    }

    if (i > 100) {
      /* Silence, don't guess */
    } else if (i > 64) {
      filehandle->sampfreq = 8000;
    } else if (i > 46) {
      filehandle->sampfreq = 11025;
    } else if (i > 32) {
      filehandle->sampfreq = 16000;
    } else if (i > 23) {
      filehandle->sampfreq = 22050;
    } else if (i > 16) {
      filehandle->sampfreq = 32000;
    } else if (i > 11) {
      filehandle->sampfreq = 44100;
    }
  }

  return 1;
}

int Audio::File::GetRawHeader(char *buf)
{
  if (fp != NULL) {
    fseek(fp, 0, SEEK_END);
    filehandle->length = (ftell(fp) - filehandle->skipBytes) / (filehandle->sampsize * filehandle->nchannels);
  }
  filehandle->headSize = filehandle->skipBytes;

  return 1;
}

int Audio::File::PutRawHeader(int len)
{
  return(0);
}


int Audio::File::GetAuHeader(char *buf)
{
  int
    tmp, hlen, 
    nsamp, nsampfile;

  if (filehandle->firstNRead < AU_HEADERSIZE) {
    if (fread((char *)&buf[filehandle->firstNRead],sizeof(char), AU_HEADERSIZE-filehandle->firstNRead, fp) < 0) {
      return -1; 
    }
  }
  hlen = GetBELong(buf, 4);
  tmp  = GetBELong(buf, 12);
  if (tmp == SND_FORMAT_MULAW_8) {
    filehandle->sampformat = MULAW;
    filehandle->sampsize = 1;
  } else if (tmp == SND_FORMAT_LINEAR_8) {
    filehandle->sampformat = LIN8OFFSET;
    filehandle->sampsize = 1;
  } else if (tmp == SND_FORMAT_LINEAR_16) {
    filehandle->sampformat = LIN16;
    filehandle->sampsize = 2;
  } else if (tmp == SND_FORMAT_ALAW_8) {
    filehandle->sampformat = ALAW;
    filehandle->sampsize = 1;
  } else {
    printf("Unsupported AU format\n");
    return -1;
  }
  
  filehandle->sampfreq = GetBELong(buf, 16);
  filehandle->nchannels = GetBELong(buf, 20);
  filehandle->headSize = hlen;
  nsamp = GetBELong(buf, 8) / (filehandle->sampsize * filehandle->nchannels);

  fseek(fp, 0, SEEK_END);
  nsampfile = (ftell(fp) - hlen) / (filehandle->sampsize * filehandle->nchannels);
  if (nsampfile < nsamp || nsamp == 0) {
    nsamp = nsampfile;
  }
  filehandle->length = nsamp;
  SwapIfLE();

  return 1;
}

int Audio::File::PutAuHeader(int len)
{
  int tmp = 0;
  char buf[HEADBUF];

  PutBELong(buf, 0, 0x2E736E64);
  PutBELong(buf, 4, AU_HEADERSIZE);
  PutBELong(buf, 8, len * filehandle->sampsize * filehandle->nchannels);

  if (filehandle->sampformat == MULAW) {
    tmp = SND_FORMAT_MULAW_8;
  } else if (filehandle->sampformat == LIN8OFFSET) {
    tmp = SND_FORMAT_LINEAR_8;
  } else if (filehandle->sampformat == LIN16) {
    tmp = SND_FORMAT_LINEAR_16;
  } else if (filehandle->sampformat == ALAW) {
    tmp = SND_FORMAT_ALAW_8;
  }
  PutBELong(buf, 12, tmp);

  PutBELong(buf, 16, filehandle->sampfreq);
  PutBELong(buf, 20, filehandle->nchannels);
  PutBELong(buf, 24, 0);

  if (fwrite(buf, sizeof(char),AU_HEADERSIZE,fp) == -1) return -1;

  if (len == -1) {
    SwapIfLE();
  }
  filehandle->inByteOrder = BIGENDIAN;

  return(AU_HEADERSIZE);
}

#ifndef WIN32
#  define WAVE_FORMAT_PCM   1
#  define WAVE_FORMAT_ALAW  6
#  define WAVE_FORMAT_MULAW 7
#endif

int Audio::File::GetHeaderBytes(char *buf, int len)
{
  int rlen = fread(&buf[filehandle->firstNRead], sizeof(char), len - filehandle->firstNRead, fp);

  if (rlen < len - filehandle->firstNRead){
    printf ("Failed reading header bytes\n");
    return -1;
  }
  filehandle->firstNRead += rlen;

  return 1;
}

int Audio::File::GetWavHeader(char *buf)
{
  int tmp, nsamp = 0, nsampfile, i = 12, chunkLen;

  /* buf[] = "RIFFxxxxWAVE" */

  while (1) {
    if (strncmp("fmt ", &buf[i], strlen("fmt ")) == 0) {
      chunkLen = GetLELong(buf, i + 4) + 8;
      if (filehandle->firstNRead < i + chunkLen) {
	if (GetHeaderBytes(buf, i + chunkLen) != 1) {
	  return -1;
	}
      }
      tmp = GetLEShort(buf, i+8);

      if (tmp == WAVE_FORMAT_PCM) {
	filehandle->sampformat = LIN16;
      } else if (tmp == WAVE_FORMAT_ALAW) {
	filehandle->sampformat = ALAW;
      } else if (tmp == WAVE_FORMAT_MULAW) {
	filehandle->sampformat = MULAW;
      } else {
	printf("Unsupported WAV format\n");
	return -1;
      }
      
      filehandle->nchannels = GetLEShort(buf, i+10);
      filehandle->sampfreq   = GetLELong(buf, i+12);
      filehandle->sampsize   = GetLEShort(buf, i+22) / 8;
      if (filehandle->sampsize == 1 && filehandle->sampformat == LIN16) {
	filehandle->sampformat = LIN8OFFSET;
      }
    } else if (strncmp("data", &buf[i], strlen("data")) == 0) {
      nsamp = GetLELong(buf, i + 4) / (filehandle->sampsize * filehandle->nchannels);
      break;
    } else { /* unknown chunk */
      chunkLen = GetLELong(buf, i + 4) + 8;
      if (chunkLen < 0 || chunkLen > HEADBUF) {
	printf("Failed parsing WAV header\n");
	return -1;
      }
      if (filehandle->firstNRead < i + chunkLen) {
	if (GetHeaderBytes(buf, i + chunkLen) != 1) {
	  return -1;
	}
      }
    }

    i += chunkLen;
    if (filehandle->firstNRead < i + 8) {
      if (GetHeaderBytes(buf, i + 8) != 1) {
	return -1;
      }
    }
    if (i >= HEADBUF) {
      printf("Failed parsing WAV header\n");
      return -1;
    }
  }
  
  filehandle->headSize = i + 8;
  fseek(fp, 0, SEEK_END);
  nsampfile = (ftell(fp) - filehandle->headSize) / (filehandle->sampsize * filehandle->nchannels);
  if (nsampfile < nsamp || nsamp == 0) {
    nsamp = nsampfile;
  }
  filehandle->length = nsamp;
  SwapIfBE();

  return 1;
}


int Audio::File::PutWavHeader(int len)
{
  char buf[HEADBUF];

  sprintf(&buf[0], "RIFF");
  if (len != -1) {
    PutLELong(buf, 4, len * filehandle->sampsize * filehandle->nchannels + 36);
  } else {
    SwapIfBE();
    PutLELong(buf, 4, 0x7FFFFFFF);
  }
  sprintf(&buf[8], "WAVEfmt ");
  PutLELong(buf, 16, 16);

  if (filehandle->sampformat == ALAW) 
    PutLEShort(buf, 20, WAVE_FORMAT_ALAW);
  else if (filehandle->sampformat == MULAW) 
    PutLEShort(buf, 20, WAVE_FORMAT_MULAW);
  else 
    PutLEShort(buf, 20, WAVE_FORMAT_PCM);

  PutLEShort(buf, 22, (short)filehandle->nchannels);
  PutLELong(buf, 24, filehandle->sampfreq);
  PutLELong(buf, 28, (filehandle->sampfreq * filehandle->nchannels * filehandle->sampsize * 8 + 7) / 8);
  PutLEShort(buf, 32, (short)((filehandle->nchannels * filehandle->sampsize * 8 + 7) / 8));
  PutLEShort(buf, 34, (short) (filehandle->sampsize * 8));
  sprintf(&buf[36], "data");
  if (len != -1) {
    PutLELong(buf, 40, len * filehandle->sampsize * filehandle->nchannels);
  } else {
    PutLELong(buf, 40, 0x7FFFFFDB);
  }
  fseek (fp, 0 , SEEK_SET);
  if (fwrite(buf, sizeof(char), WAV_HEADERSIZE, fp) == -1) return -1;
  filehandle->inByteOrder = LITTLEENDIAN;

  return(WAV_HEADERSIZE);
}

/* See http://www.borg.com/~jglatt/tech/aiff.htm */

unsigned long Audio::File::ConvertFloat(unsigned char *buffer)
{
  unsigned long mantissa;
  unsigned long last = 0;
  unsigned char exp;
  
  memcpy(&mantissa, buffer + 2, sizeof(long));
#ifdef LE
  mantissa = SwapLong(mantissa);
#endif
  exp = 30 - *(buffer+1);
  while (exp--) {
    last = mantissa;
    mantissa >>= 1;
  }
  if (last & 0x00000001) mantissa++;
  return(mantissa);
}

void Audio::File::StoreFloat(unsigned char * buffer, unsigned long value)
{
  
  unsigned long exp;
  unsigned char i;
  
  exp = value;
  exp >>= 1;
  for (i=0; i<32; i++) {
    exp >>= 1;
    if (!exp) break;
  }
  *buffer = (char)64;
  *(buffer+1) = i;
  
  for (i=32; i; i--) {
    if (value & 0x80000000) break;
    value <<= 1;
  }


#ifdef LE
  value = SwapLong(value);
#endif  
  memcpy(buffer + 2, &value, sizeof(long));
}

int Audio::File::GetAiffHeader(char *buf)
{
  int tmp, i = 12, chunkLen = 4;

  /* buf[] = "FORMxxxxAIFF" */

  while (1) {
    if (strncmp("COMM", &buf[i], strlen("COMM")) == 0) {
      chunkLen = GetBELong(buf, i + 4) + 8;
      if (filehandle->firstNRead < i + chunkLen) {
	if (GetHeaderBytes(buf, i + chunkLen) != 1) {
	  return -1;
	}
      }
      filehandle->nchannels = GetBEShort(buf, i + 8);
      tmp = GetBEShort(buf, i + 14);
      if (tmp == 16) {
	filehandle->sampformat = LIN16;
	filehandle->sampsize = 2;
      } else {
	filehandle->sampformat = LIN8;
	filehandle->sampsize = 1;
      }
      filehandle->sampfreq = ConvertFloat((unsigned char *)&buf[i+16]);
    } else if (strncmp("SSND", &buf[i], strlen("SSND")) == 0) {
      chunkLen = 16;
      if (filehandle->firstNRead < i + chunkLen) {
	if (GetHeaderBytes(buf, i + 8) != 1) {
	  return -1;
	}
      }
      filehandle->length = (GetBELong(buf, i + 4) - 8) / (filehandle->sampsize * filehandle->nchannels);
      tmp = GetBELong(buf, i + 8);
      break;
    } else {
      chunkLen = 4;
      if (i > HEADBUF - 4) {
	printf("Missing chunk in AIFF header\n");
	return -1;
      }
    }
    i += chunkLen;
    if (filehandle->firstNRead < i + 8) {
      if (GetHeaderBytes(buf, i + 8) != 1) {
	return -1;
      }
    }
  }
  filehandle->headSize = i + 12 + tmp;
  SwapIfLE();

  return 1;
}


int Audio::File::PutAiffHeader(int len)
{
  char buf[HEADBUF];

  
  sprintf(&buf[0], "FORM");
  if (len != -1) {
    PutBELong(buf, 4, len * filehandle->sampsize * filehandle->nchannels + AIFF_HEADERSIZE);
  } else {
    PutBELong(buf, 4, 0x7FFFFFFF);
  }
  sprintf(&buf[8], "AIFF");

  sprintf(&buf[12], "COMM");
  PutBELong(buf, 16, 18);
  PutBEShort(buf, 20, (short) filehandle->nchannels);
  PutBELong(buf, 22, len);
  PutBEShort(buf, 26, (short) (filehandle->sampsize * 8));
  StoreFloat((unsigned char *) &buf[28], (long) filehandle->sampfreq);
  sprintf(&buf[38], "SSND");
  if (len != -1) {
    PutBELong(buf, 42, (len * ((filehandle->sampsize + 7) / 8) * filehandle->nchannels)); // +8
  } else {
    PutBELong(buf, 42, 8 + 0x7FFFFFD1);
  }
  PutBELong(buf, 46, 0);
  PutBELong(buf, 50, 0);
  fseek (fp, 0 , SEEK_SET);
  if (fwrite(buf, sizeof(char), AIFF_HEADERSIZE, fp) == -1) return -1;
 
  filehandle->inByteOrder = BIGENDIAN;

  return(AIFF_HEADERSIZE);
}
