/* 
 * Copyright (C) 1997-2000 Kare Sjolander <kare@speech.kth.se>
 *
 * This file is part of the Snack sound extension for Tcl/Tk.
 * The latest version can be found at http://www.speech.kth.se/snack/
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <glob.h>

#ifdef __NetBSD__
 #include <soundcard.h>
#else /* Linux */
 #include <sys/soundcard.h>
#endif

#include "AudioPort.h"
#include "AudioFile.h"
#include "g711.h"


#ifndef min
#define min(a,b) ((a)<(b)?(a):(b))
#define max(a,b) ((a)>(b)?(a):(b))
#endif

int mfd = 0;

double Time::Current()
{
#ifdef WIN32
  struct timeb t;
  
  ftime(&t);

  return(t.time + t.millitm * 0.001);
#else
  struct timeval tv;
  struct timezone tz;
  
  (void) gettimeofday(&tv, &tz);

  return(tv.tv_sec + tv.tv_usec * 0.000001);

#endif
}

char *String::Duplicate(const char *str) {
  char *newStr;

  newStr = (char *)malloc(strlen(str)+1);

  if (newStr) {
    strncpy(newStr, str, strlen(str)+1);
  }

  return newStr;
}

Audio::Port::Port()
{
  defaultDeviceName = DEVICE_NAME;
}

Audio::Port::~Port()
{
}

int Audio::Port::Open(int _mode, int _freq, int _channels, int _encoding, char *_device)
{
  char
    *device;
  int 
    format,
    nformat,
    channels,
    speed,
    mask,
    frag,
    i, j;


  if (_device == NULL) {
    device = defaultDeviceName;
  }
  else if (strlen(device) == 0) {
    device = defaultDeviceName;
  }
  else {
    device = _device;
  }

  desc.mode = _mode;

  switch (_mode) {
  case RECORD:
    if ((desc.afd = open(device, O_RDONLY)) == -1) {
//       Tcl_AppendResult(interp, "Could not open ", device, " for read.",
// 		       NULL);
      return -1;
    }
    break;

  case PLAY:
    if ((desc.afd = open(device, O_WRONLY, 644)) == -1) {
//       Tcl_AppendResult(interp, "Could not open ", device, " for write.",
// 		       NULL);
      return -1;
    } 
    j = 1;
    for (i=0; (j*2)<=BUFSIZE; i++) {
      j = j*2;
    }
    switch (i) {
    case 7:
      frag = 0x7fff0007;
      break;
    case 8:
      frag = 0x7fff0008;
      break;
    case 9:
      frag = 0x7fff0009;
      break;
    case 10:
      frag = 0x7fff0010;
      break;
    default:
      frag = 0x7fff0009;
      break;
    }

    if (ioctl(desc.afd, SNDCTL_DSP_SETFRAGMENT, &frag) < 0) {
      fprintf(stderr,"ERROR: cant't set fragsize\n");
      return -1;
    }
    break;
  }

  fcntl(desc.afd, F_SETFD, FD_CLOEXEC);

  ioctl(desc.afd, SNDCTL_DSP_GETFMTS, &mask);

  desc.convert = 0;

  switch (_encoding) {
  case LIN16:
    format = AFMT_S16_LE;
    desc.bytesPerSample = sizeof(short);
    break;
  case ALAW:
    if (mask & AFMT_A_LAW) {
      format = AFMT_A_LAW;
      desc.bytesPerSample = sizeof(char);
    } 
    else {
      format = AFMT_S16_LE;
      desc.bytesPerSample = sizeof(short);
      desc.convert = ALAW;
    }
    break;
  case MULAW:
    if (mask & AFMT_MU_LAW) {
      format = AFMT_MU_LAW;
      desc.bytesPerSample = sizeof(char);
    }
    else {
      format = AFMT_S16_LE;
      desc.bytesPerSample = sizeof(short);
      desc.convert = MULAW;
    }
    break;
  case LIN8OFFSET:
    format = AFMT_U8;
    desc.bytesPerSample = sizeof(char);
    break;
  case LIN8:
    format = AFMT_S8;
    desc.bytesPerSample = sizeof(char);
    break;
  }

  nformat = format;
  if (ioctl(desc.afd, SNDCTL_DSP_SETFMT, &format) == -1
      || format != nformat) {
    close(desc.afd);
//     Tcl_AppendResult(interp, "Failed setting format.", NULL);
    return -1;
  }

  desc.nChannels = _channels;
  channels = _channels;
  if (ioctl(desc.afd, SNDCTL_DSP_CHANNELS, &channels) == -1
      || channels != _channels) {
    close(desc.afd);
//     Tcl_AppendResult(interp, "Failed setting number of channels.", NULL);
    return -1;
  }

  speed = _freq;
  if (ioctl(desc.afd, SNDCTL_DSP_SPEED, &speed) == -1
      || abs(speed - _freq) > _freq / 100) {
    close(desc.afd);
//     Tcl_AppendResult(interp, "Failed setting sample frequency.", NULL);
    return -1;
  }

  desc.count = 0;
  desc.time  = Time::Current();
  desc.timep = 0.0;
  desc.freq  = _freq;

  return 1;
}

int Audio::Port::Close()
{
  desc.count = 0;
  close(desc.afd);
  desc.afd = 0;

  return(0);
}

int Audio::Port::Pause()
{
  int 
    res;

  res = Played();
  desc.timep = Time::Current();
  ioctl(desc.afd, SNDCTL_DSP_RESET, 0);

  return(res);
}

void Audio::Port::Resume()
{
  desc.time = desc.time + Time::Current() - desc.timep;
}

void Audio::Port::Flush()
{
  if (desc.mode == RECORD) {
  } 
  else {
    ioctl(desc.afd, SNDCTL_DSP_RESET, 0);
  }
}


void Audio::Port::Post()
{
  ioctl(desc.afd, SNDCTL_DSP_POST, 0);
}

int Audio::Port::Read(void *buf, int nFrames)
{
  int 
    n, i, 
    res;
  short 
    s[2];

  if (desc.convert) {
    n = 0;
    for (i = 0; i < nFrames * desc.nChannels; i += desc.nChannels) {
      res = read(desc.afd, &s, desc.nChannels * sizeof(short));
      if (res <= 0) return(n / (desc.bytesPerSample * desc.nChannels));
      if (desc.convert == ALAW) {
	((unsigned char *)buf)[i] = G711::Lin2Alaw(s[0]);
	if (desc.nChannels == 2) {
	  ((unsigned char *)buf)[i+1] = G711::Lin2Alaw(s[1]);
	}
      } else {
	((unsigned char *)buf)[i] = G711::Lin2Mulaw(s[0]);
	if (desc.nChannels == 2) {
	  ((unsigned char *)buf)[i+1] = G711::Lin2Mulaw(s[1]);
	}
      }
      n += res;
    }

    return(n / (desc.bytesPerSample * desc.nChannels));
  } 
  else {
    n = read(desc.afd, buf, nFrames * desc.bytesPerSample * desc.nChannels);
    if (n > 0) n /= (desc.bytesPerSample * desc.nChannels);

    return(n);
  }
}

int Audio::Port::Write(void *buf, int nFrames)
{
  int 
    n, i, 
    res;
  short 
    s;

  if (desc.convert) {
    n = 0;
    for (i = 0; i < nFrames * desc.nChannels; i++) {
      if (desc.convert == ALAW) {
	s = G711::Alaw2Lin(((unsigned char *)buf)[i]);
      } else {
	s = G711::Mulaw2Lin(((unsigned char *)buf)[i]);
      }
      res = write(desc.afd, &s, sizeof(short));
      if (res <= 0) 
	return(n / (desc.bytesPerSample * desc.nChannels));
      n += res;
    }

    return(n / (desc.bytesPerSample * desc.nChannels));
  } 
  else {
    n = write(desc.afd, buf, nFrames * desc.bytesPerSample * desc.nChannels); 
    if (n > 0) n /= (desc.bytesPerSample * desc.nChannels);
    
    return(n);
  }
}

int Audio::Port::Readable()
{
  audio_buf_info 
    info;

  ioctl(desc.afd, SNDCTL_DSP_GETISPACE, &info);
  return (info.bytes / (desc.bytesPerSample * desc.nChannels));
}

int Audio::Port::Writeable()
{
  audio_buf_info 
    info;

  ioctl(desc.afd, SNDCTL_DSP_GETOSPACE, &info);

  return (info.bytes / (desc.bytesPerSample * desc.nChannels));
}

int Audio::Port::Played()
{
  /*
    count_info info;
    
    ioctl(A->afd, SNDCTL_DSP_GETOPTR, &info);
    
    return(A->count + (info.bytes / (A->bytesPerSample * A->nChannels)) - 
    A->frag_size);
    */
  int res;
  
  res = (desc.freq * (Time::Current() - desc.time) +.5);
  return(res);
}

int Audio::Port::GetDevices(char **arr, int n)
{
  int 
    i, j;
  glob_t 
    globt;
  
  i = 0;
  j = 0;

  glob("/dev/dsp*", 0, NULL, &globt);
  glob("/dev/audio*", GLOB_APPEND, NULL, &globt);

  for (i = 0; i < globt.gl_pathc; i++) {
    if (j < n) {
      arr[j++] = (char *) String::Duplicate(globt.gl_pathv[i]);
    }
  }
  globfree(&globt);

  return(j);
}

void Audio::Port::GetFormats(char *buf, int n)
{
  int 
    afd,
    mask, 
    pos;

  if (desc.afd == 0) {
    if ((afd = open(DEVICE_NAME, O_WRONLY, 0)) == -1) {
      buf[0] = '\0';
      return;
    }
  }
  else
    afd = desc.afd;

  if (ioctl(afd, SNDCTL_DSP_GETFMTS, &mask) == -1) {
    buf[0] = '\0';
    return;
  }

  pos = 0;
  if (mask & AFMT_S16_LE) {
    pos += sprintf(&buf[pos], "%s", "Lin16");
  }
  if (mask & AFMT_U8) {
    pos += sprintf(&buf[pos], "%s", " Lin8offset");
  }
  if (mask & AFMT_S8) {
    pos += sprintf(&buf[pos], "%s", " Lin8");
  }
  pos += sprintf(&buf[pos], "%s", " Mulaw Alaw");

  if (desc.afd == 0) {
    close(afd);
  }
}

void Audio::Port::GetFrequencies(char *buf, int n)
{
  int 
    afd,
    freq, 
    pos, i;
  int 
    f[] = { 8000, 11025, 16000, 22050, 32000, 44100, 48000 };

  if (desc.afd == 0) {
    if ((afd = open(DEVICE_NAME, O_WRONLY, 0)) == -1) {
      buf[0] = '\0';
      return;
    }
  }
  else
    afd = desc.afd;

  pos = 0;
  for (i = 0; i < 7; i++) {
    freq = f[i];
    if (ioctl(afd, SNDCTL_DSP_SPEED, &freq) == -1) break;
    if (abs(f[i] - freq) > freq / 100) continue;
    pos += sprintf(&buf[pos], "%d ", freq);
  }

  if (desc.afd == 0) {
    close(afd);
  }
}


/*??*/


